/*
 * StreamJsonReader.cpp
 *
 *  Created on: 07/03/2014
 *      Author: laso
 *
 *  Modified on: 07/26/2015
 *      Author: xopr
 */

#include "StreamJsonReader.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>

//#define DEBUG

#ifdef DEBUG
#if defined(ARDUINO) && ARDUINO >= 100
    #include "Arduino.h"
#elif defined(SPARK_CORE)
    #include "Application.h"
#endif
#endif

/* A C++ program to implement itoa() */
//#include <iostream>
//using namespace std;

// http://www.geeksforgeeks.org/implement-itoa/
// A utility function to reverse a string
void reverse(char str[], int length)
{
    int start = 0;
    int end = length -1;
    char temp;
    while (start < end)
    {
        temp = *(str+start);
        *(str+start) = *(str+end);
        *(str+end) = temp;
        //swap(*(str+start), *(str+end));
        start++;
        end--;
    }
}
 
// Implementation of itoa()
char* itoa_ex(int num, char* str, int base)
{
    int i = 0;
    bool isNegative = false;
 
    // Handle 0 explicitely, otherwise empty string is printed for 0
    if (num == 0)
    {
        str[i++] = '0';
        str[i] = '\0';
        return str;
    }
 
    // In standard itoa(), negative numbers are handled only with 
    // base 10. Otherwise numbers are considered unsigned.
    if (num < 0 && base == 10)
    {
        isNegative = true;
        num = -num;
    }
 
    // Process individual digits
    while (num != 0)
    {
        int rem = num % base;
        str[i++] = (rem > 9)? (rem-10) + 'a' : rem + '0';
        num = num/base;
    }
 
    // If number is negative, append '-'
    if (isNegative)
        str[i++] = '-';
 
    str[i] = '\0'; // Append string terminator
 
    // Reverse the string
    reverse(str, i);
 
    return str;
}

StreamJsonReader::StreamJsonReader(const char* queries[],
								  unsigned int num_queries,
								  unsigned int max_name_size,
								  unsigned int max_value_size,
								  unsigned int max_trace_size) : queries((char**)queries),
																 num_queries(num_queries),
																 max_value_size(max_value_size),
																 max_trace_size(max_trace_size),
																 max_name_size(max_name_size),
																 realloc_increment(5),
																 found_results(0),
																 ignore_node(false){


	alloc_elements();
}

StreamJsonReader::~StreamJsonReader(){
	free_elements();
}

void StreamJsonReader::free_elements(){
	free(this->element_value);
	free(this->element_name);
	free(this->trace);
	for(int i=0; i<num_queries; i++){
		free(this->results[i]);
	}
	free(this->results);
}

void StreamJsonReader::alloc_elements(){
	this->trace = (char*) malloc(max_trace_size * sizeof(char));
	this->element_name = (char*) malloc(max_name_size * sizeof(char));
	this->element_value = (char*) malloc(max_value_size * sizeof(char));


	status = NODE_VALUE;
	element_type = TYPE_DICT;
	element_value[0] = '\0';
	element_name[0] = '\0';
	trace[0] = '\0';

	//Alloc memory for result pointers
	results = (char**) malloc(num_queries * sizeof(char*));

	//Init results with a \0
	for(int i=0; i<num_queries; i++){
		results[i] = (char*) malloc(1 * sizeof(char));
		results[i][0] = '\0';
	}
}

void StreamJsonReader::reset(){
	free_elements();
	found_results = 0;
	status = NODE_VALUE;
	alloc_elements();
}

// MATCH PARTIAL

void StreamJsonReader::assign_result(char* result){
	//Check trace against queries and assign result
	for(int i=0; i < num_queries; i++){
		if(query_match(queries[i], trace)){
			results[i] = (char*) realloc(results[i], (strlen(result)+1)*sizeof(char));
			strcpy(results[i], result);
			found_results++;
			return;
		}
	}
}

bool StreamJsonReader::finished(){
	return found_results == num_queries;
}

bool StreamJsonReader::query_match(char* query, char* trace){
	return strcmp( query, trace ) == 0;
}

bool StreamJsonReader::partial_query_match(char* trace){
	//query starts with trace, we need to keep looking on this node
	for(int i=0; i < num_queries; i++){
		if ( strncmp( queries[i], trace, strlen( trace ) ) )
		{
			return true;
		}
	}
	return false;
}



bool StreamJsonReader::is_numeric_last_trace_element(){
	int size = strlen(trace);
	if(size==0){
		return false;
	}
	for(int i=size-1; i >= 0; i--){
		if(trace[i] == '.'){
			return true;
		}
		if(!(trace[i] >= '0' && trace[i] <= '9')){
			return false;
		}
	}
	return true;
}

int StreamJsonReader::numeric_last_trace_element_value(){
	//cosa.1  => 1
	int size = strlen(trace);
	int value = 0;
	if(!is_numeric_last_trace_element()){
		return -1;
	}
	for(int i=size-1; i >= 0; i--){
		if(trace[i] != '.'){
			value += atoi(trace + i ) * pow(10, (size-1-i));
		}
		else{
			return value;
		}
	}
	return -1;
}


void StreamJsonReader::increment_trace_element(){
	//cosa.1 to cosa.2
	int value = numeric_last_trace_element_value();
	if(value == -1){
		return;
	}
	int size = strlen(trace);
	char as_string[5]; // limited to 5 numbers indexes
	//int_to_string(value + 1, as_string);
itoa_ex( value + 1, as_string, 10 );

	for(int i=size-1; i >= 0; i--){
		if(trace[i] == '.'){
			for(int j=0; j <= strlen(as_string); j++){ //copy \0 too
				if(i+j+2 < max_trace_size){
					trace[i+1+j] = as_string[j];
				}
			}
			return;
		}
	}
	return;
}

void StreamJsonReader::remove_last_trace_element(){
	int size = strlen(trace);
	int i;
	for(i=size-1; i >= 0; i--){
		if(trace[i] == '.'){
			trace[i] = '\0';
			ignore_node = !partial_query_match(trace);
			return;
		}
	}
	trace[0] = '\0';
	ignore_node = !partial_query_match(trace);

}

void StreamJsonReader::add_to_trace(char* element){

	if(strlen(element) == 0){
		return;
	}
	int size = strlen(trace);
	if(size > 0){
		if(!append_to_trace('.')){
			return;
		}
		size++;
	}
	for(int i=0; i < strlen(element); i++){ //copy \0
		append_to_trace(element[i]);
	}
	ignore_node = !partial_query_match(trace);
}

int StreamJsonReader::process_char(char c)
{
#ifdef DEBUG
    Serial.print( "s:" );
    Serial.print( status, DEC );
    Serial.print( "t:" );
    Serial.print( element_type, DEC );
    Serial.println( "" );
#endif

	switch( status )
	{
		case NODE_VALUE: //At the beggining is a node value of root. Wait for { or [ or number or "
			switch( c )
			{
                // Ignore whitespace and newlines
                case ' ':
                case '\t':
                case '\n':
                case '\r':
                    break;

				case '{': status = READ_ELEMENT; element_type = TYPE_DICT; break;
				case '[': status = NODE_VALUE; add_to_trace((char*) "0"); break;
				case '"': status = READ_ELEMENT; element_type = TYPE_STRING; element_value[0] = '\0'; break;
				case ']': remove_last_trace_element(); status = WAIT_FOR_ANOTHER_ELEMENT_OR_CLOSE; break; //Empty array
             
                case 't': // true
                case 'f': // false
                case 'n': // null
                    status = READ_ELEMENT;
                    element_type = TYPE_SPECIAL;
                    append_to_value(c); //must be cleared before here
                    break;

				default:
					if((c >= '0' && c <= '9') || c == '-'){
						 status = READ_ELEMENT;
						 element_type = TYPE_NUMERIC;
						 append_to_value(c); //must be cleared before here
					}
					else{
						return -1;
					}
					break;
			}
			break;
		case READ_ELEMENT: //We are reading an element of type element_type
			switch( element_type )
			{
				case TYPE_DICT:
					// Inside a dict we only expect a element name
					if(c == '}')
					{
						add_to_trace((char*) "0");
						status = WAIT_FOR_ANOTHER_ELEMENT_OR_CLOSE;
						return process_char(c);
					}
                   
                    // Ignore whitespace and newlines
                    if ( c == ' ' || c == '\t' || c == '\n' || c == '\r' ) 
                        break;

					if ( c != '"' )
                        return -1;

					status = READING_ELEMENT_KEY;
					element_name[0] = '\0';
					break;

				case TYPE_STRING:
					// Inside a string, concat until "
					if(c != '"') append_to_value(c);
					else{
						//cout << "*PUT STRING! " << element_value << " ==> " << trace << endl;
						assign_result(element_value);
						element_value[0] = '\0';
						status = WAIT_FOR_ANOTHER_ELEMENT_OR_CLOSE;
					}
					break;

      case TYPE_SPECIAL:
				case TYPE_NUMERIC:
					// Inside a numeric, concat until } or ,
					if(c != '}' && c != ',' && c != ']') append_to_value(c);
					else{
						// FIN VALOR DE NUMERO!!! DEVOLVERLO!!
						//cout << "*PUT NUMERO! " << element_value << " ==> " << trace << endl;
						assign_result(element_value);
						element_value[0] = '\0';
						status = WAIT_FOR_ANOTHER_ELEMENT_OR_CLOSE;
						return process_char(c); // Advance one status
					}
			}//end switch
			break;
		case WAIT_FOR_ANOTHER_ELEMENT_OR_CLOSE:

			if(c == '}' || c == ']'){
				status = WAIT_FOR_ANOTHER_ELEMENT_OR_CLOSE;
				remove_last_trace_element(); // parent value ended
				if( is_numeric_last_trace_element( ) )
      {
        
					//element_type = TYPE_ARRAY;
				}
				else{
					element_type = TYPE_DICT;
				}
			}
			else if(c == ','){
				if(!is_numeric_last_trace_element()){
					status = READ_ELEMENT;
					remove_last_trace_element(); // value with key ended
					element_type = TYPE_DICT;
				}
				else{
					status = NODE_VALUE; // Any type of value
					increment_trace_element();
					//element_type = TYPE_ARRAY;
				}
			}

			break;
		case READING_ELEMENT_KEY:
			if(c != '"'){
				append_to_name(c);
			}
			else{
				status = WAIT_FOR_TWO_POINTS; // !!!!!! READ FOR MATCH NAME!!
				add_to_trace(element_name);
				element_name[0] = '\0';
			}
			break;

		case WAIT_FOR_TWO_POINTS:
			if(c != ':') return -1;
			else status = NODE_VALUE;
			break;
	}

	return 0;
}

bool StreamJsonReader::append_to_trace(char c){
	int size = strlen(this->trace);
	if(size + 1 > this->max_trace_size){
		this->trace = resize_string(this->trace, &this->max_trace_size); //Modifies string pointer if realloc moves on memory
	}

this->trace[ size++ ] = c;
this->trace[ size ] = 0;
/*
	if(!append_to_string(this->trace, c, this->max_trace_size)){
		return false;
	}
*/
	return true;
}

bool StreamJsonReader::append_to_value( char c ){
	int size = strlen(this->element_value);
	if(size + 1 > this->max_value_size){
		this->element_value = resize_string(this->element_value, &this->max_value_size); //Modifies string pointer if realloc moves on memory
	}

this->element_value[ size++ ] = c;
this->element_value[ size ] = 0;
/*
	if(!append_to_string(this->element_value, c, this->max_value_size)){
		return false;
	}
	*/
	return true;
}

bool StreamJsonReader::append_to_name(char c){
	int size = strlen(this->element_name);
	if(size + 1 > this->max_name_size){
		this->element_name = resize_string(this->element_name, &this->max_name_size); //Modifies string pointer if realloc moves on memory
	}
if ( size >= this->max_name_size )
  return false;

this->element_name[ size++ ] = c;
this->element_name[ size ] = 0;
/*
	if(!append_to_string(this->element_name, c, this->max_name_size)){

		return false;
	}
	*/
	return true;
}


unsigned int* StreamJsonReader::get_string_max_size(char* string){
	unsigned int *max_size;
	if(string==this->element_name){
		max_size = &this->max_name_size;
	}
	else if(string==this->element_value){
		max_size = &this->max_value_size;
	}
	else if(string==this->trace){
		max_size = &this->max_trace_size;
	}
	return max_size;
}


char* StreamJsonReader::resize_string(char* string, unsigned int* max_size){
	size_t next_size = (size_t) (*max_size + realloc_increment) * sizeof(char);

	char* tmp = (char*) realloc(string, next_size);
	if(tmp == NULL){
		return string;
	}
	else{
		*max_size += realloc_increment;
		return tmp;
	}
}


int atoi(char c){
	return c - '0';
}

